//Windows header
#ifdef WIN32
	#include <windows.h>
#endif

#include <cmath>
#ifndef M_PI
	#define M_PI 3.14159265358979323846f
#endif

// OpenGL header
#include <GL/gl.h>

#include <stdlib.h>
#include <time.h>
#include <iostream>
using namespace std;


#include "vec.h"
#include "dynamics.h"
#include "objLoader.h"
#include "animation.h"


/////////////////////////////////////////
//  Animation methods
////////////////////////////////////////

void Animation::setDt() {
	// compute elapsed time
	float newTime = (float)clock()/(float)CLOCKS_PER_SEC;
	dt = (newTime-oldTime);
	oldTime = newTime;
}

void Animation::animateDynamics() {
	updateDynamicsParameters();
	// update visual model :
	if (clothMesh) clothMesh->updateMesh(dynamics.positions);
}

void Animation::animateSkinning() {
	skinning.animate(dt);
}

void Animation::updateDynamicsParameters()
{
	dt = 0.0025;

	// compute dynamics :
	dynamics.animate(dt);
}

void Animation::updateCollisionModelFromSkinning(glm::mat4 *m_transfo) {
	// For each sphere of character :
	for (unsigned int sp = indexCharacter ; sp < dynamics.getNbBalls() ; sp++) {
		// Get initial position :
		Vec v = initPosCollisionModel[sp-indexCharacter];
		// Get joint influencing :
		unsigned int influ = influences[sp-indexCharacter];
		// Get transfo :
		glm::mat4 transfo = m_transfo[influ];
		//////////////////////////////////////////
		//////////// Question TP2-3 : ////////////
		//////////////////////////////////////////
		// Apply transfo to sphere :
		
	}
}

////////////////////////////////////////
// Initialization methods
////////////////////////////////////////

void Animation::initScene(unsigned int nfpp)
{
	clothMesh = new Mesh();
	characterMesh = new Mesh();

	// init dynamics and skinning :
	setDynamics();
	skinning.num_influences = nfpp;
	setSkinning();
}

void Animation::setDynamics()
{
	// Load mechanical and collision model :
	dynamics.setFromMesh("data/cloth.obj", clothMesh);
	indexCharacter = dynamics.getNbBalls();
	dynamics.addSphereModel("data/girlSphere.txt");
	nbBalls = dynamics.getNbBalls();
}

void Animation::setSkinning() {
	// Load skin :
	objLoader character;
	character.load("data/girl.obj");
	characterMesh->points = character.points;
	characterMesh->normals = character.normals;
	characterMesh->faces = character.faces;
	characterMesh->triangles = character.triangles;
	characterMesh->nbEdges = character.nbEdges;
	characterMesh->color=Point(0.75,0.25,0,1);

	skinning.initModelData(characterMesh->points.size());

	setInfluences();
}

void Animation::setInfluences() {
	FILE *fdat = fopen("data/girlSphere.txt", "r");
	char line[300];
	float x,y,z,r; // (x,y,z,inf) = position of center of sphere, r = radius; 
	int inf; // inf = bone influencing sphere
	while(fgets(line, 300, fdat)) { // for each sphere
		if (sscanf(line, "%f %f %f %f %d", &x,&y,&z,&r,&inf)==5) {
			initPosCollisionModel.push_back(Vec(x,y,z));
			influences.push_back(inf);
		}
	}
	fclose(fdat);
}

///////////////////////////////////////////////
//  Draw methods
//////////////////////////////////////////////

void Animation::drawGround()
{
	float backColor[] = {1.0, 0.0, 0.0};
	float frontColor[] = {0.0, 0.0, 1.0};
	
	glBegin(GL_QUADS);
		glColor3fv(frontColor);
		glVertex3f(-10.0f, 0.0f, -10.0f);
		glColor3fv(backColor);
		glVertex3f(-10.0f, 0.0f,  10.0f);
		glColor3fv(backColor);
		glVertex3f( 10.0f, 0.0f,  10.0f);
		glColor3fv(frontColor);
		glVertex3f( 10.0f, 0.0f, -10.0f);	
	glEnd();
}



void Animation::drawSphere()
{
	unsigned int nbStrips = 5;
	float stepVAngle = (float)M_PI/nbStrips;
	
	unsigned int nbFacesPerStrip = 10;
	float stepHAngle = 2*(float)M_PI/nbFacesPerStrip;

	for(unsigned int i=0 ; i<nbStrips ; i++) {
		float vAngle1 = i*stepVAngle - 0.5f*(float)M_PI;
		float vAngle2 = (i+1)*stepVAngle - 0.5f*(float)M_PI;
		
		// draw one strip
		float intensity1 = (float)i/nbStrips;
		float intensity2 = (float)(i+1)/nbStrips;
		glBegin(GL_QUAD_STRIP);
			for (unsigned int j=0 ; j<=nbFacesPerStrip ; j++) {
				float hAngle = j*stepHAngle;						
				float pos1[] = {cos(hAngle)*cos(vAngle1), sin(vAngle1), sin(hAngle)*cos(vAngle1)};
				float pos2[] = {cos(hAngle)*cos(vAngle2), sin(vAngle2), sin(hAngle)*cos(vAngle2)};
				
				glNormal3fv(pos1);
				glColor3f(intensity1, 0.0f, 1.0f-intensity1);
				glVertex3fv(pos1);
				
				glNormal3fv(pos2);
				glColor3f(intensity2, 0.0f, 1.0f-intensity2);
				glVertex3fv(pos2);
			}
		glEnd();
	}
}


void Animation::drawBalls(bool drawMass)
{
	for (unsigned int i=0 ; i<nbBalls ; i++){
		if (!drawMass && !dynamics.forceDraw[i]) continue;
		glPushMatrix();
		
		Vec pos = dynamics.getBallPos(i);
		glTranslatef(pos.x, pos.y, pos.z);

		double radius = dynamics.getBallRadius(i);
		glScaled(radius, radius, radius);
		
		drawSphere();
		
		glPopMatrix();
	}
}

void Animation::drawSprings()
{
	unsigned int nbSprings = dynamics.getNbSprings();
	glBegin(GL_LINES);
	for (unsigned int i=0 ; i<nbSprings ; i++){
		unsigned int obj1 = dynamics.getSpring(i).obj1;
		unsigned int obj2 = dynamics.getSpring(i).obj2;
		Vec pos1 = dynamics.getBallPos(obj1);
		Vec pos2 = dynamics.getBallPos(obj2);
		glColor3f(1.0f, 1.0f, 1.0f);
		glVertex3f(pos1.x, pos1.y, pos1.z);
		glVertex3f(pos2.x, pos2.y, pos2.z);
	}		
	glEnd();
}


void Animation::renderScene(void) 
{
	// draw objects
	drawGround();
	// mechanical and collision model :
	if (drawCollisionModel) drawBalls(drawMass); 
	if (drawMass) drawSprings();
	// visual model :
	if (clothMesh && drawVisualModel) clothMesh->draw();
	//if (characterMesh && drawVisualModel) characterMesh->draw();
}