#include "skinning.h"

#include <cmath>
#ifndef M_PI
	#define M_PI 3.14159265358979323846f
#endif

#include <stdlib.h>
#include <time.h>
#include <iostream>

#include "vec.h"

void Skinning::initModelData(unsigned int nv) {
	num_vertices = nv;

	char objName[120];
	FILE *file;
	char line[100];

	// Read positions of bones :
	sprintf(objName, "data/jointPos.txt");
	file = fopen(objName, "r");
	while (!feof(file)) {
		if (fgets(line, 100, file)) {
			float x,y,z;
			if (sscanf(line, "%f %f %f", &x, &y, &z)) {
				jointsInitPos.push_back(Vec3(x,y,z));
			}
		}		
	}
	fclose(file);

	// Read orientations of bones :
	sprintf(objName, "data/jointOrient.txt");
	file = fopen(objName, "r");
	while (!feof(file)) {
		if (fgets(line, 100, file)) {
			float x,y,z,w;
			if (sscanf(line, "%f %f %f %f", &x, &y, &z, &w)) {
				jointsInitAxis.push_back(Vec3(x,y,z));
				jointsInitAngle.push_back(w);
			}
		}		
	}
	fclose(file);

	jointsPos.resize(jointsInitPos.size());
	jointsAxis.resize(jointsInitAxis.size());
	jointsAngle.resize(jointsInitAngle.size());

	num_bones = (unsigned int)jointsInitPos.size();

	// Create hierarchy :
	parent.push_back(-1);
	parent.push_back(0);
	parent.push_back(1);
	parent.push_back(2);
	parent.push_back(1);
	parent.push_back(4);
	parent.push_back(5);
	parent.push_back(6);
	parent.push_back(0);
	parent.push_back(8);
	parent.push_back(9); 
	parent.push_back(10);
	parent.push_back(1);
	parent.push_back(12);
	parent.push_back(13);
	parent.push_back(14);
	parent.push_back(0);
	parent.push_back(16);
	parent.push_back(17); 
	parent.push_back(18);


	// Read skinning info (index, weight) :
	std::vector<float> bone_indexA;
	std::vector<float> bone_weightA;
	sprintf(objName, "data/skinning.txt");
	file = fopen(objName, "r");
	char * pch;
	while (!feof(file)) {
		if (fgets(line, 100, file)) {
			float x;
			float count = 0;
			int c=0;
			pch = strtok(line," ");
			while (pch != NULL) {
				if (pch[0]=='\n'){
				} else {
					x = (float)atof(pch);
					bone_weightA.push_back(x);
					bone_indexA.push_back(count);
					if (x!=0) c++;
					count++;
				}
				pch = strtok (NULL, " ");
			}
		}		
	}
	fclose(file);

	bone_index.resize(num_influences*num_vertices,0);
	bone_weight.resize(num_influences*num_vertices,0);
	unsigned nfpp = num_influences; // number of frames per point
    for( unsigned int x=0; x < num_vertices; x++){
		int bi = 0;
		for (unsigned int b = 0 ; b < num_bones ; b++) {
			if (bone_weightA[num_bones*x+b]!=0) {
				bone_index[nfpp*x+bi] = bone_indexA[num_bones*x+b];
				bone_weight[nfpp*x+bi] = bone_weightA[num_bones*x+b];
				bi++;
			}
		}
    }

	dofs.resize(num_bones, Vec3(0,0,0));

	updateJoints();
}

// Conversion (axis, angle) -> matrix
void axis2mat(glm::mat3 *R, const Vec3 axis, const float angle) {
	Vec v(axis);
	v.normalize();
	float x = v.x*sin(angle/2.0);
	float y = v.y*sin(angle/2.0);
	float z = v.z*sin(angle/2.0);
	float s = cos(angle/2.0);
	float x2 = x*x;
	float y2 = y*y;
	float z2 = z*z;
	float xy = x*y;
	float xz = x*z;
	float yz = y*z;
	float sx = s*x;
	float sy = s*y;
	float sz = s*z;
	
	(*R)[0][0] = 1.0 - 2.0 * (y2 + z2);
    (*R)[0][1] = 2.0 * (xy - sz);
    (*R)[0][2] = 2.0 * (xz + sy);

    (*R)[1][0] = 2.0 * (xy + sz);
    (*R)[1][1] = 1.0 - 2.0 * (z2 + x2);
    (*R)[1][2] = 2.0 * (yz - sx);

    (*R)[2][0] = 2.0 * (xz - sy);
    (*R)[2][1] = 2.0 * (yz + sx);
    (*R)[2][2] = 1.0 - 2.0 * (y2 + x2);
}

// Conversion matrix -> (axis, angle)
void mat2axis(Vec3 *axis, float *angle, const glm::mat3 R) {
	float saux = (R[0][0]+R[1][1]+R[2][2]-1)/2.0f;
	if (saux < -1) saux=-1;
	if (saux > 1) saux=1;
	float s = acos(saux);
	if (s<1e-9) {
		(*angle)=0;
		(*axis).x = 1;
		(*axis).y = 0;
		(*axis).z = 0;
	} else {
		(*angle)=s;
		Vec v;
		v.x = (R[2][1]-R[1][2])/(2.0*sin(*angle));
		v.y = (R[0][2]-R[2][0])/(2.0*sin(*angle));
		v.z = (R[1][0]-R[0][1])/(2.0*sin(*angle));
		v.normalize();
		axis->x = v.x;
		axis->y = v.y;
		axis->z = v.z;
	}
}

void Skinning::animate(float dt) {
	// Set degrees of freedom :
	dofs[5].y+=0.5; if (dofs[5].y>120) dofs[5].y=120;

	globalTra.z+=0.01;

	// Update the bones current position/orientation accordingly :
	updateJoints();
}

void Skinning::updateJoints() {
	glm::mat3 R0dof;
	glm::mat3 Rdof;
	float a;
	// Update pos and angle/axis :
	for (unsigned int x = 0 ; x < num_bones ; x++) {
		// Translation : none except if joint 0 :
		jointsPos[x] = jointsInitPos[x];
		if (x==0) {
			jointsPos[x]+=globalTra;
		}

		// Rotation :
		// Initial one :
		Vec3 axis = jointsInitAxis[x];
		float angle = jointsInitAngle[x];
		angle *= M_PI/180.0;

		R0dof = glm::mat3(1.0f);

		// rx :
		a = M_PI*(float)(dofs[x].x)/(float)180.0;
		axis2mat(&Rdof, Vec3(1,0,0), a);
		R0dof = Rdof*R0dof;
		
		// ry :
		a = M_PI*(float)(dofs[x].y)/(float)180.0;
		axis2mat(&Rdof, Vec3(0,1,0), a);
		R0dof = Rdof*R0dof;
		
		// rz :
		a = M_PI*(float)(dofs[x].z)/(float)180.0;
		axis2mat(&Rdof, Vec3(0,0,1), a);
		R0dof = Rdof*R0dof;

		axis2mat(&Rdof, axis, angle);
		R0dof=Rdof*R0dof;
		
		mat2axis(&axis, &angle, R0dof);

		jointsAngle[x] = 180.0*angle/M_PI;
		jointsAxis[x] = axis;
	}
}